//
//  AMRequestQueue.m
//  AMFatFractal
//
//  Created by Andre Muis on 8/26/13.
//  Copyright (c) 2013 Andre Muis. All rights reserved.
//

#import "AMRequestQueue.h"

#import "AMCreateRequest.h"
#import "AMDeleteRequest.h"
#import "AMGetRequest.h"
#import "AMGetArrayRequest.h"
#import "AMLoginRequest.h"
#import "AMEvents.h"
#import "AMEvent.h"
#import "AMFatFractal.h"
#import "AMRequest.h"
#import "AMUpdateRequest.h"

@interface AMRequestQueue ()

@property (strong, nonatomic, readonly) AMFatFractal *amFatFractal;
@property (strong, nonatomic) NSMutableArray *queue;
@property (strong, nonatomic, readonly) AMEvents *events;
@property (assign, nonatomic, readonly) float retryDelaySeconds;

@end

@implementation AMRequestQueue

- (id)initWithAMFatFractal: (AMFatFractal *)amFatFractal retryDelaySeconds: (float)retryDelaySeconds
{
    self = [super init];
    
    if (self)
    {
        _amFatFractal = amFatFractal;
        _queue = [[NSMutableArray alloc] init];
        _events = [[AMEvents alloc] init];
        _retryDelaySeconds = retryDelaySeconds;
    }
    
    return self;
}

- (int)count
{
    return self.queue.count;
}

- (void)add: (AMRequest *)request
{
    [self.queue addObject: request];
}

- (void)addThenProcess: (AMRequest *)request
{
    [self add: request];
    [self process];
}

- (void)process
{
    for (AMRequest *request in self.queue)
    {
        if ([request isKindOfClass: [AMLoginRequest class]] == YES)
        {
            AMLoginRequest *loginRequest = (AMLoginRequest *)request;
            
            [self.amFatFractal loginWithUserName: loginRequest.userName
                                     andPassword: loginRequest.password
                                      onComplete: ^(NSError *theErr, id theObj, NSHTTPURLResponse *theResponse)
             {
                 if (loginRequest.executeCompletionHandler == YES)
                 {
                     loginRequest.completionHandler(theErr, theObj, theResponse);
                 }
                 
                 if (theErr != nil || theResponse.statusCode != 200)
                 {
                     loginRequest.retryEvent = [[AMEvent alloc] initWithTarget: self
                                                                        action: @selector(addThenProcess:)
                                                                        object: loginRequest
                                                                  intervalSecs: self.retryDelaySeconds
                                                                       repeats: NO];
                     
                     [self.events addEvent: loginRequest.retryEvent];
                     
                     loginRequest.retryCount = loginRequest.retryCount + 1;
                 }
             }];
        }
        else if ([request isKindOfClass: [AMGetRequest class]] == YES)
        {
            AMGetRequest *getRequest = (AMGetRequest *)request;
            
            [self.amFatFractal getObjFromUri: getRequest.uri
                                  onComplete: ^(NSError *theErr, id theObj, NSHTTPURLResponse *theResponse)
             {
                 if (getRequest.executeCompletionHandler == YES)
                 {
                     getRequest.completionHandler(theErr, theObj, theResponse);
                 }
                 
                 if (theErr != nil || theResponse.statusCode != 200)
                 {
                     getRequest.retryEvent = [[AMEvent alloc] initWithTarget: self
                                                                      action: @selector(addThenProcess:)
                                                                      object: getRequest
                                                                intervalSecs: self.retryDelaySeconds
                                                                     repeats: NO];
                     
                     [self.events addEvent: getRequest.retryEvent];
                     
                     getRequest.retryCount = getRequest.retryCount + 1;
                 }
             }];
        }
        else if ([request isKindOfClass: [AMGetArrayRequest class]] == YES)
        {
            AMGetArrayRequest *getArrayRequest = (AMGetArrayRequest *)request;
            
            [self.amFatFractal getArrayFromUrl: getArrayRequest.uri
                                    onComplete: ^(NSError *theErr, id theObj, NSHTTPURLResponse *theResponse)
             {
                 if (getArrayRequest.executeCompletionHandler == YES)
                 {
                     getArrayRequest.completionHandler(theErr, theObj, theResponse);
                 }
                 
                 if (theErr != nil || theResponse.statusCode != 200)
                 {
                     getArrayRequest.retryEvent = [[AMEvent alloc] initWithTarget: self
                                                                           action: @selector(addThenProcess:)
                                                                           object: getArrayRequest
                                                                     intervalSecs: self.retryDelaySeconds
                                                                          repeats: NO];
                     
                     [self.events addEvent: getArrayRequest.retryEvent];
                     
                     getArrayRequest.retryCount = getArrayRequest.retryCount + 1;
                 }
             }];
        }
        else if ([request isKindOfClass: [AMCreateRequest class]] == YES)
        {
            AMCreateRequest *createRequest = (AMCreateRequest *)request;
            
            [self.amFatFractal createObj: createRequest.object
                                   atUri: createRequest.uri
                              onComplete: ^(NSError *theErr, id theObj, NSHTTPURLResponse *theResponse)
            {
                if (createRequest.executeCompletionHandler == YES)
                {
                    createRequest.completionHandler(theErr, theObj, theResponse);
                }

                if (theErr != nil || theResponse.statusCode != 201)
                {
                    createRequest.retryEvent = [[AMEvent alloc] initWithTarget: self
                                                                        action: @selector(addThenProcess:)
                                                                        object: createRequest
                                                                  intervalSecs: self.retryDelaySeconds
                                                                       repeats: NO];
                    
                    [self.events addEvent: createRequest.retryEvent];
                    
                    createRequest.retryCount = createRequest.retryCount + 1;
                }
            }];
        }
        else if ([request isKindOfClass: [AMUpdateRequest class]] == YES)
        {
            AMUpdateRequest *updateRequest = (AMUpdateRequest *)request;
            
            [self.amFatFractal updateObj: updateRequest.object
                              onComplete: ^(NSError *theErr, id theObj, NSHTTPURLResponse *theResponse)
             {
                 if (updateRequest.executeCompletionHandler == YES)
                 {
                     updateRequest.completionHandler(theErr, theObj, theResponse);
                 }
                 
                 if (theErr != nil || theResponse.statusCode != 200)
                 {
                     updateRequest.retryEvent = [[AMEvent alloc] initWithTarget: self
                                                                         action: @selector(addThenProcess:)
                                                                         object: updateRequest
                                                                   intervalSecs: self.retryDelaySeconds
                                                                        repeats: NO];
                     
                     [self.events addEvent: updateRequest.retryEvent];
                     
                     updateRequest.retryCount = updateRequest.retryCount + 1;
                 }
             }];
        }
        else if ([request isKindOfClass: [AMDeleteRequest class]] == YES)
        {
            AMDeleteRequest *deleteRequest = (AMDeleteRequest *)request;
            
            [self.amFatFractal deleteObj: deleteRequest.object
                              onComplete: ^(NSError *theErr, id theObj, NSHTTPURLResponse *theResponse)
             {
                 if (deleteRequest.executeCompletionHandler == YES)
                 {
                     deleteRequest.completionHandler(theErr, theObj, theResponse);
                 }
                 
                 if (theErr != nil || theResponse.statusCode != 200)
                 {
                     deleteRequest.retryEvent = [[AMEvent alloc] initWithTarget: self
                                                                         action: @selector(addThenProcess:)
                                                                         object: deleteRequest
                                                                   intervalSecs: self.retryDelaySeconds
                                                                        repeats: NO];
                     
                     [self.events addEvent: deleteRequest.retryEvent];
                     
                     deleteRequest.retryCount = deleteRequest.retryCount + 1;
                 }
             }];
        }
    }
    
    [self.queue removeAllObjects];
}

- (void)cancel: (AMRequest *)request
{
    [self.events removeEvent: request.retryEvent];
    request.executeCompletionHandler = NO;
}

@end


























